/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package actor.components;

import graphics.geometry.Geometry;
import graphics.shader.Effect;

import java.util.LinkedList;
import java.util.List;

import logging.Logger;
import mesh.BoundingSphere;
import mesh.Material;
import mesh.Mesh;
import mesh.MeshLoader;

import org.lwjgl.util.vector.Vector3f;
import org.w3c.dom.Node;

import util.GeometryFactory;
import util.XMLUtil;
import actor.ActorComponent;

/**
 * Ein Actor mit einer RenderComponent kann vom Graphicslayer gezeichnet werden.
 * @author nico3000
 */
public class RenderComponent extends ActorComponent {
    /*
     * Die ID der Komponente.
     */
    public static final int ID = 0x5a263f25;
    
    /**
     * Liste aller Geometrien.
     */
    private final List<Mesh> meshes = new LinkedList<>();
    /**
     * Die BoundingSphere des Meshes in Modelkoordinaten ohne Translation.
     */
    private BoundingSphere boundingSphere = null;
    /**
     * Der Effect, der fuer diesen Actor benutzt werden soll.
     */
    private Effect effect = null;

    @Override
    protected int getComponentId() {
        return ID;
    }

    @Override
    protected boolean init(Node data) {
        while(data != null) {
            switch(data.getNodeName()) {
                case "SimpleCuboid":
                    this.addCuboid(data);
                    break;
                case "SimpleSphere":
                    this.addSphere(data);
                    break;
                case "ObjMesh":
                    this.addObjMesh(data);
                    break;
                case "Effect":
                    String vs = XMLUtil.getAttributeAsString(data, "vs");
                    String fs = XMLUtil.getAttributeAsString(data, "fs");
                    if(vs == null || fs == null) {
                        Logger.INSTANCE.error("Effect tag needs vertex AND fragment shader!");
                    } else {
                        Effect customEffect = new Effect();
                        if(customEffect.initVF(vs, fs)) {
                            this.effect = customEffect;
                            Material.modifyEffectToUseMaterial(this.effect);
                        }
                    }
                    break;
                case "BoundingSphere":
                    Vector3f center = XMLUtil.getAttributesAsVector(XMLUtil.getChildByType(data, "Center"), "x", "y", "z", null);
                    float radius = XMLUtil.getAttributeAsFloat(XMLUtil.getChildByType(data, "Radius"), "value", 0.0f);
                    this.boundingSphere = new BoundingSphere();
                    this.boundingSphere.init(center, radius);
                    //this.meshes.add(MeshLoader.createMesh(GeometryFactory.createSphere(32, radius, center.x, center.y, center.z)));
                    break;
                case "#comment":
                    // nothing to do
                    // no break
                case "#text":
                    // nothing to do
                    break;
                default:
                    Logger.INSTANCE.warning("node \"" + data.getNodeName() + "\" was ignored");
            }            
            data = data.getNextSibling();
        }
        return true;
    }

    @Override
    protected void destroy() {
        if(this.effect != null) {
            this.effect.destroy();
        }
        for(Mesh mesh : this.meshes) {
            mesh.delRef();
        }
    }
    
    /**
     * Bindet und zeichnet alle Geometrien mit dem aktuell benutzten Effekt
     * oder, sofern vorhanden, mit einem eigenen.
     */
    public void drawGeometries() {
        Effect oldEffect = Effect.getCurrentEffect();
        if(this.effect != null) {
            this.effect.use();
        }
        for(Mesh mesh : this.meshes) {
            mesh.draw();
        }
        oldEffect.use();
    }

    /**
     * Liefert die BoundingSphere in Modelkoordinaten.
     * @return Die BoundingSphere
     */
    public BoundingSphere getBoundingSphere() {
        return this.boundingSphere;
    }
    
    /**
     * Fuegt einen simplen Quader hinzu.
     * @param data Der <tt>SimpleCuboid</tt> Node
     */
    private void addCuboid(Node couboidNode) {
        Vector3f size = XMLUtil.getAttributesAsVector(XMLUtil.getChildByType(couboidNode, "Size"), "x", "y", "z", null);
        Vector3f pos = XMLUtil.getAttributesAsVector(XMLUtil.getChildByType(couboidNode, "Position"), "x", "y", "z", null);
        Geometry geo = GeometryFactory.createCuboid(size.x, size.y, size.z, pos.x, pos.y, pos.z);
        if(geo != null) {
            this.meshes.add(MeshLoader.createMesh(geo));
        }
    }
    
    /**
     * Fuegt eine simple Kugel hinzu.
     * @param sphereNode Der <tt>SimpleSphere</tt> Node
     */
    private void addSphere(Node sphereNode) {
        float radius = XMLUtil.getAttributeAsFloat(XMLUtil.getChildByType(sphereNode, "Size"), "radius", 0.0f);
        int subDivision = XMLUtil.getAttributeAsInt(XMLUtil.getChildByType(sphereNode, "Size"), "sub", 32);
        Vector3f pos = XMLUtil.getAttributesAsVector(XMLUtil.getChildByType(sphereNode, "Position"), "x", "y", "z", null);
        Geometry geo = GeometryFactory.createSphere(subDivision, radius, pos.x, pos.y, pos.z);
        if(geo != null) {
            this.meshes.add(MeshLoader.createMesh(geo));
        }
    }
    
    /**
     * Fuegt ein Mesh aus einer obj-Datei hinzu.
     * @param meshNode Der <tt>ObjMesh</tt> Node
     */
    private void addObjMesh(Node meshNode) {
        String resource = XMLUtil.getAttributeAsString(meshNode, "resource");
        boolean mirrored = XMLUtil.getAttributeAsBoolean(meshNode, "mirrored", false);
        String customMaterial = XMLUtil.getAttributeAsString(meshNode, "customMaterialLibrary");
        if(resource != null) {
            Mesh mesh = MeshLoader.loadObjMesh(resource, customMaterial, mirrored);
            if(mesh != null) {
                this.meshes.add(mesh);
            }
        }
    }
}
