/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package graphics.geometry;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL30;

/**
 * Eine Geometrie ist eine Veknuepfung von einem Index- und einem Vertexbuffer.
 * @author nico3000
 */
public class Geometry {
    /**
     * Die ID der Vertex array obejct.
     */
    private int id = 0;
    /**
     * Reference Counter.
     */
    private int ref = 0;
    /**
     * Der Indexbuffer.
     */
    private IndexBuffer indexBuffer = null;
    /**
     * Der Vertexbuffer.
     */
    private VertexBuffer vertexBuffer = null;
    
    /**
     * Initialisiert die Geometrie mit ihrem Index- und Vertexbuffer.
     * @param indexBuffer Der Indexbuffer
     * @param vertexBuffer Der Vertexnuffer
     * @return <tt>true</tt>, falls eroglreich
     */
    public boolean init(IndexBuffer indexBuffer, VertexBuffer vertexBuffer) {
        this.indexBuffer = indexBuffer;
        this.vertexBuffer = vertexBuffer;
        this.id = GL30.glGenVertexArrays();
        if(this.id < 1) {
            return false;
        } else {
            this.bind();
            this.indexBuffer.bind();
            this.vertexBuffer.bind();
            this.vertexBuffer.bindAttributes();
            GL30.glBindVertexArray(0);
            return true;
        }
    }
    
    /**
     * Bindet die Geometrie und kann anschliessend mit <tt>draw()</tt> 
     * gezeichnet werden.
     */
    public void bind() {
        GL30.glBindVertexArray(this.id);
    }
    
    /**
     * Erhoeht den Reference Counter um 1.
     */
    public void addRef() {
        ++this.ref;
    }
    
    /**
     * Verringert den Reference Counter um 1. Ist er danach kleiner oder gleich
     * 0, wird die Geometrie zerstoert.
     */
    public void delRef() {
        if(--this.ref <= 0) {
            if(this.vertexBuffer != null) {
                this.vertexBuffer.destroy();
                this.vertexBuffer = null;
            }
            if(this.indexBuffer != null) {
                this.indexBuffer.destroy();
                this.indexBuffer = null;
            }
            if(this.id > 0) {
                GL30.glDeleteVertexArrays(this.id);
            }
        }
    }
    
    /**
     * Zeichnet die Geometrie. Dazu muss sie vorher mit <tt>bind()</tt> gebunden
     * werden.
     */
    public void render() {
        GL11.glDrawElements(this.indexBuffer.getTopology(), this.indexBuffer.getCount(), GL11.GL_UNSIGNED_INT, 0);
    }
    
    /**
     * Zeichnet einen Teil der Geometrie. Dazu muss sie vorher mit
     * <tt>bind()</tt> gebunden werden.
     */
    public void draw(int startIndex, int indexCount) {
        GL11.glDrawElements(this.indexBuffer.getTopology(), indexCount, GL11.GL_UNSIGNED_INT, 4 * startIndex);
    }
}
