/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package graphics.texture;

import java.nio.FloatBuffer;
import logging.Logger;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL30;

/**
 * Eine Textur ist eine Sammlung von Daten auf der Grafikkarte, die lesenden
 * wahlfreien Zugriff im Shader ermoeglichen.
 * @author nico3000
 */
public class Texture {    
    /**
     * Unterstuetzte Typen von Texturen.
     */
    public static enum Type {
        /**
         * Eindimensionale Textur.
         */
        TEXTURE_1D,
        /**
         * Zweidimensionale Textur.
         */
        TEXTURE_2D,
    }
    
    /**
     * Die Slots und die momentan daran gebundenen Texturen.
     */
    private static final Texture bindings[] = new Texture[80];
    /**
     * Der als Working Slot definierte Slot.
     */
    private static final int WORKING_SLOT = 0;
    /**
     * Der momentan aktive Slot.
     */
    private static int activeSlot = -1;
    
    /**
     * Die ID der Textur.
     */
    private int id = 0;
    /**
     * Das Binding Target der Textur.
     */
    private final int target;
    /**
     * Reference Counter.
     */
    private int ref = 0;
    /**
     * Groesse der Textur im VRAM in Byte.
     */
    private int size = 0;

    /**
     * Konstruktor.
     * @param type Der Typ der Textur
     */
    public Texture(Type type) {
        switch(type) {
            case TEXTURE_1D:
                this.target = GL11.GL_TEXTURE_1D;
                break;
            case TEXTURE_2D:
                this.target = GL11.GL_TEXTURE_2D;
                break;
            default:
                // should not happen
                this.target = 0;
                break;
        }
    }
    
    /**
     * Initialisiert die Textur.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        this.destroy();
        this.id = GL11.glGenTextures();
        return this.id > 0;
    }
    
    /**
     * Zerstoert die Textur und ihre Daten.
     */
    private void destroy() {
        if(this.id != 0) {
            GL11.glDeleteTextures(this.id);
            this.id = 0;
        }
    }
    
    public void addRef() {
        ++this.ref;
    }
    
    public void delRef() {
        if(--this.ref == 0) {
            this.destroy();
        }
    }

    /**
     * Liefert die OpenGL Konstante des Targets zurueck.
     * @return Das Target
     */
    public int getTarget() {
        return target;
    }
    
    /**
     * Bindet die Textur an ihr Target.
     */
    public void bind() {
        activateSlot(WORKING_SLOT);
        GL11.glBindTexture(this.target, this.id);
    }
    
    /**
     * Verbindet diese Textur mit ihrem Slot, falls sie es noch nicht ist.
     */
    public void connectToSlot(int slot) {
        if(bindings[slot] != this) {
            activateSlot(slot);
            GL11.glBindTexture(this.target, this.id);
            bindings[slot] = this;
        }
    }
    
    /**
     * Aktiviert einen Slot, falls er nicht schon aktiv ist.
     * @param slot Der zu aktivierende Slot
     */
    private static void activateSlot(int slot) {
        if(slot != activeSlot) {
            GL13.glActiveTexture(GL13.GL_TEXTURE0 + slot);
            activeSlot = slot;
        }
    }
    
    /**
     * Setzt die Daten der Textur.
     * @param width Die Breite
     * @param height Die Hoehe
     * @param channels Die Anzahl der Farbkanaele
     * @param data Die Daten
     */
    public void setData(int width, int height, int channels, FloatBuffer data) {
        int format;
        switch(channels) {
            case 1: format = GL11.GL_RED; break;
            case 2: format = GL30.GL_RG; break;
            case 3: format = GL11.GL_RGB; break;
            case 4: format = GL11.GL_RGBA; break;
            default:
                Logger.INSTANCE.error("unsupported color channel count");
                return;
        }        
        this.bind();
        GL11.glTexImage2D(this.target, 0, GL11.GL_RGBA8, width, height, 0, format, GL11.GL_FLOAT, data);
        GL11.glTexParameteri(this.target, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR_MIPMAP_LINEAR);
        GL11.glTexParameteri(this.target, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
        GL30.glGenerateMipmap(this.target);
        
        this.size = 0;
        int maxLevelSize = width * height * channels;
        int dim = Math.max(width, height);
        while(dim != 0) {
            this.size += maxLevelSize;
            maxLevelSize /= 4;
            dim /= 2;
        }
    }

    public int getSize() {
        return size;
    }
}
