/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh;

import java.nio.FloatBuffer;
import java.util.List;

import logging.Logger;

import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

import util.MathUtil;

/**
 *
 * @author nico3000
 */
public class BoundingSphere {
    private final Vector3f currentCenter = new Vector3f();
    private final Vector3f initialCenter = new Vector3f();
    private float radius = 0.0f;
    
    public void init(Vector3f initialCenter, float radius) {
        this.initialCenter.set(initialCenter);
        this.radius = radius;        
        this.currentCenter.set(initialCenter);
    }
    
    public void transform(Matrix4f model) {
        MathUtil.transformCoord(this.currentCenter, model, this.initialCenter);
    }

    public Vector3f getInitialCenter() {
        return this.initialCenter;
    }

    public float getRadius() {
        return this.radius;
    }
    
    public boolean isCompleteOnInvisibleSide(Vector4f plane) {
        float distance = plane.x * this.currentCenter.x + plane.y * this.currentCenter.y + plane.z * this.currentCenter.z + plane.w;
        return distance + this.radius < 0.0f;
    }
    
    public void encapsulateSpheres(List<BoundingSphere> spheres) {
        Vector3f sum = new Vector3f();
        for(BoundingSphere sphere : spheres) {
            Vector3f.add(sum, sphere.currentCenter, sum);
        }
        this.initialCenter.set(sum.x / (float)spheres.size(), sum.y / (float)spheres.size(), sum.z / (float)spheres.size());
        this.currentCenter.set(this.initialCenter);
        this.radius = 0.0f;
        for(BoundingSphere sphere : spheres) {
            this.radius = Math.max(this.radius, Vector3f.sub(sphere.currentCenter, this.initialCenter, sum).length() + sphere.radius);
        }
    }
    
    public void encapsulateVertices(FloatBuffer vertices, int vertexStride, int posOffset) {
        int vertexCount = vertices.capacity() / vertexStride;
        final Vector3f p1 = new Vector3f(), p2 = new Vector3f(), temp = new Vector3f();
        final Vector3f a = new Vector3f(), b = new Vector3f();
        float distance = 0.0f;
        for(int i=0; i < vertexCount; ++i) {
            this.loadPosition(p1, i, vertices, vertexStride, posOffset);
            for(int j=i+1; j < vertexCount; ++j) {
                this.loadPosition(p2, j, vertices, vertexStride, posOffset);
                if(Vector3f.sub(p2, p1, temp).lengthSquared() > distance) {
                    a.set(p1);
                    b.set(p2);
                    distance = temp.lengthSquared();
                }
            }
        }
        MathUtil.lerp(this.initialCenter, a, b, 0.5f);
        for(int i=0; i < vertexCount; ++i) {
            distance = Math.max(distance, Vector3f.sub(p1, this.initialCenter, temp).lengthSquared());
        }
        this.radius = 0.5f * MathUtil.sqrt(distance);
    }
    
    private void loadPosition(Vector3f dest, int vertexNr, FloatBuffer vertices, int vertexStride, int posOffset) {
        vertices.position(vertexNr * vertexStride + posOffset);
        dest.load(vertices);
    }

    @Override
    public String toString() {
        return "BoundingSphere{" + "currentCenter=" + currentCenter + ", radius=" + radius + '}';
    }
}
