/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh;

import graphics.geometry.Geometry;
import java.util.HashMap;
import java.util.Map;
import logging.Logger;

/**
 * Der Meshloader ist die zentrale Anlaufstelle, um Meshes zu erzeugen.
 * @author nico3000
 */
public class MeshLoader {
    /**
     * Ein Platzhaltermesh, dass anstelle von noch nicht geladenen Meshes
     * angezeigt wird.
     */
    public static final Mesh DEFAULT_MESH = new Mesh("meshes/exclamation.obj");
    /**
     * Map aller geladenen Meshes, indiziert nach ihrer Resource.
     */
    private static final Map<String, Mesh> meshes = new HashMap<>();
    
    /**
     * Initialisiert das <tt>DEFAULT_MESH</tt>.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public static boolean init() {
        if(!DEFAULT_MESH.initMesh(false)) {
            Logger.INSTANCE.error("default mesh could not be initialized");
            return false;
        } else {
            DEFAULT_MESH.addRef();
            return true;
        }
    }
    
    /**
     * Gibt das <tt>DEFAULT_MESH</tt> wieder frei.
     */
    public static void destroy() {
        DEFAULT_MESH.delRef();
    }
    
    /**
     * Laedt ein Mesh aus einer obj-Resource und initialisiert es. Wurde aus der
     * Resource schon einmal ein Mesh erzeugt und ist dieses Mesh noch
     * vorhanden, so wird dieses verwendet.
     * @param resource Die obj-Resource
     * @param mirrored Soll das Objekt gespiegelt werden?
     * @return Das Mesh, falls erfolgreich initialisiert; sonst <tt>null</tt>
     */
    public static Mesh loadObjMesh(String resource, String material, boolean mirrored) {
        //TODO Mesh kopieren, wenn es ein neues Material zugewiesen bekommt
    	
    	Mesh mesh = null;
    	
    	if (material != null && !material.equals("")) {
    		mesh = meshes.get(resource + material);
    		if (mesh != null) {
    			mesh.addRef();
    			return mesh;
    		}
    	}
    	if (mesh == null) {
    		mesh = meshes.get(resource);
    	}
    	
        if(mesh == null) {
            mesh = new Mesh(resource);
            if(!mesh.initMesh(mirrored)) {
                Logger.INSTANCE.error("mesh could not be initialized");
                return null;
            }
            //Logger.INSTANCE.info(mesh.getStats());
            meshes.put(resource, mesh);
        }

        if (material != null && !material.equals("")) {
        	Mesh customMesh = new Mesh(mesh);
        	customMesh.setMtlLib(material);
        	meshes.put(resource + material, customMesh);
        	customMesh.addRef();
        	return customMesh;
        }
        
        mesh.addRef();
        return mesh;
    }

    /**
     * Erzeugt ein Mesh aus einer Geometry.
     * @param geometry Die Geometrie
     * @return Das Mesh
     */
    public static Mesh createMesh(Geometry geometry) {
        Mesh mesh = new Mesh(geometry);
        mesh.addRef();
        return mesh;
    }
}
