/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package process;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import logging.Logger;

/**
 * Verwaltet und aktualisiert registrierte Prozesse.
 * @author nico3000
 */
public class ProcessManager {
    /**
     * Alle registrierten Prozesse.
     */
    private final LinkedList<EngineProcess> processes = new LinkedList<>();
    
    /**
     * Temporaere Liste fuer neue Prozesse.
     */
    private final List<EngineProcess> tempProcesses = new LinkedList<>();
    
    /**
     * Fuegt einen Prozess hinzu.
     * @param process Der neue Prozess
     */
    public void attachProcess(EngineProcess process) {
        this.tempProcesses.add(process);
    }
    
    /**
     * Aktualisiert alle Prozesse.
     * @param deltaMillis Dauer des letzten Frames in Milliskunden (Gametimer)
     * @return Die Anzahl der erfolgreich beendeten und gescheiterten Prozesse
     */
    public int update(long deltaMillis) {
        this.processes.addAll(this.tempProcesses);
        this.tempProcesses.clear();
        int succeeded = 0;
        int failed = 0;
        ListIterator<EngineProcess> iter = this.processes.listIterator();
        for(EngineProcess process; iter.hasNext();) {
            process = iter.next();
            if(process.getState() == EngineProcess.State.UNINITIALIZED) {
                process.onInit();
            } else if(process.getState() == EngineProcess.State.RUNNING) {
                process.onUpdate(deltaMillis);
            } else if(process.isDead()) {
                switch(process.getState()) {
                    case ABORTED:
                        process.onAbort();
                        iter.remove();
                        ++failed;
                        break;
                    case FAILED:
                        process.onFail();
                        iter.remove();
                        ++failed;
                        break;
                    case SUCCEEDED:
                        process.onSuccess();
                        if(process.getChild() != null) {
                            iter.set(process.getChild());
                        } else {
                            iter.remove();
                            ++succeeded;
                        }
                        break;
                }
            }
        }
        return (failed << 16) | succeeded;
    }
    
    /**
     * Extrahiert die Anzahl der erfolgreich beendeten Prozesse aus dem
     * Rueckgabewert von <tt>update()</tt>.
     * @param returnOfUpdate Rueckgabewert von <tt>updateProcesses()</tt>
     * @return Anzahl der erfolgreich beendeten Prozesse
     */
    public int getSucceeded(int returnOfUpdate) {
        return returnOfUpdate & 0x0000ffff;
    }
    
    /**
     * Extrahiert die Anzahl der gescheiterten Prozesse aus dem Rueckgabewert
     * von <tt>update()</tt>.
     * @param returnOfUpdate Rueckgabewert von <tt>updateProcesses()</tt>
     * @return Anzahl der gescheiterten Prozesse
     */
    public int getFailed(int returnOfUpdate) {
        return (returnOfUpdate & 0xffff0000) >> 16;
    }
    
    /**
     * Bricht alle Prozesse ab.
     * @param immedeate Wenn true, werden die Prozesse sofort benachrichtigt;
     *                  andernfalls erst beim naechsten Aufruf von
     *                  <tt>update()</tt>
     */
    public void abortAllProcesses(boolean immedeate) {
        ListIterator<EngineProcess> iter = this.processes.listIterator();
        while(iter.hasNext()) {
            EngineProcess process = iter.next();
            process.setState(EngineProcess.State.ABORTED);
            if(immedeate) {
                process.onAbort();
                iter.remove();
            }
        }
    }
}
