/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package util;

import graphics.geometry.Geometry;
import graphics.geometry.IndexBuffer;
import graphics.geometry.VertexBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import mesh.IndexBufferInterval;
import mesh.Material;
import mesh.Mesh;
import mesh.obj.ObjData;
import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;

/**
 * Erzeugt haeufig benutzte Geometrien.
 * @author nico3000
 */
public class GeometryFactory {
	
	private static final Map<String, Geometry> geometries = new HashMap<>();
	
    /**
     * Erzeugt einen Quader.
     * @param sizeX Kantenlaenge in x-Richtung
     * @param sizeY Kantenlaenge in y-Richtung
     * @param sizeZ Kantenlaenge in z-Richtung
     * @param posX Koordinate des Mittelpunkts
     * @param posY Koordinate des Mittelpunkts
     * @param posZ Koordinate des Mittelpunkts
     * @return Der Quader
     */
    public static Geometry createCuboid(float sizeX, float sizeY, float sizeZ, float posX, float posY, float posZ) {
        Vector3f positions[] = {
            new Vector3f(posX - 0.5f * sizeX, posY - 0.5f * sizeY, posZ - 0.5f * sizeZ),
            new Vector3f(posX + 0.5f * sizeX, posY - 0.5f * sizeY, posZ - 0.5f * sizeZ),
            new Vector3f(posX + 0.5f * sizeX, posY + 0.5f * sizeY, posZ - 0.5f * sizeZ),
            new Vector3f(posX - 0.5f * sizeX, posY + 0.5f * sizeY, posZ - 0.5f * sizeZ),
            new Vector3f(posX - 0.5f * sizeX, posY - 0.5f * sizeY, posZ + 0.5f * sizeZ),
            new Vector3f(posX + 0.5f * sizeX, posY - 0.5f * sizeY, posZ + 0.5f * sizeZ),
            new Vector3f(posX + 0.5f * sizeX, posY + 0.5f * sizeY, posZ + 0.5f * sizeZ),
            new Vector3f(posX - 0.5f * sizeX, posY + 0.5f * sizeY, posZ + 0.5f * sizeZ),
        };
        FloatBuffer vertices = BufferUtils.createFloatBuffer(8 * 4 * 6);
        IntBuffer indices = BufferUtils.createIntBuffer(5 * 6);
        // 0123 -z
        Vector3f normal = new Vector3f(0.0f, 0.0f, -1.0f);
        positions[0].store(vertices); normal.store(vertices); new Vector2f(0.0f, 0.0f).store(vertices);
        positions[1].store(vertices); normal.store(vertices); new Vector2f(1.0f, 0.0f).store(vertices);
        positions[3].store(vertices); normal.store(vertices); new Vector2f(0.0f, 1.0f).store(vertices);
        positions[2].store(vertices); normal.store(vertices); new Vector2f(1.0f, 1.0f).store(vertices);
        indices.put(new int[] { 0, 1, 2, 3, IndexBuffer.RESTART });
        // 4037 -x
        normal = new Vector3f(-1.0f, 0.0f, 0.0f);
        positions[4].store(vertices); normal.store(vertices); new Vector2f(0.0f, 0.0f).store(vertices);
        positions[0].store(vertices); normal.store(vertices); new Vector2f(1.0f, 0.0f).store(vertices);
        positions[7].store(vertices); normal.store(vertices); new Vector2f(0.0f, 1.0f).store(vertices);
        positions[3].store(vertices); normal.store(vertices); new Vector2f(1.0f, 1.0f).store(vertices);
        indices.put(new int[] { 4, 5, 6, 7, IndexBuffer.RESTART });
        // 5476 +z
        normal = new Vector3f(0.0f, 0.0f, 1.0f);
        positions[5].store(vertices); normal.store(vertices); new Vector2f(0.0f, 0.0f).store(vertices);
        positions[4].store(vertices); normal.store(vertices); new Vector2f(1.0f, 0.0f).store(vertices);
        positions[6].store(vertices); normal.store(vertices); new Vector2f(0.0f, 1.0f).store(vertices);
        positions[7].store(vertices); normal.store(vertices); new Vector2f(1.0f, 1.0f).store(vertices);
        indices.put(new int[] { 8, 9, 10, 11, IndexBuffer.RESTART });
        // 1562 +x
        normal = new Vector3f(1.0f, 0.0f, 0.0f);
        positions[1].store(vertices); normal.store(vertices); new Vector2f(0.0f, 0.0f).store(vertices);
        positions[5].store(vertices); normal.store(vertices); new Vector2f(1.0f, 0.0f).store(vertices);
        positions[2].store(vertices); normal.store(vertices); new Vector2f(0.0f, 1.0f).store(vertices);
        positions[6].store(vertices); normal.store(vertices); new Vector2f(1.0f, 1.0f).store(vertices);
        indices.put(new int[] { 12, 13, 14, 15, IndexBuffer.RESTART });
        // 4510 -y
        normal = new Vector3f(0.0f, -1.0f, 0.0f);
        positions[4].store(vertices); normal.store(vertices); new Vector2f(0.0f, 0.0f).store(vertices);
        positions[5].store(vertices); normal.store(vertices); new Vector2f(1.0f, 0.0f).store(vertices);
        positions[0].store(vertices); normal.store(vertices); new Vector2f(0.0f, 1.0f).store(vertices);
        positions[1].store(vertices); normal.store(vertices); new Vector2f(1.0f, 1.0f).store(vertices);
        indices.put(new int[] { 16, 17, 18, 19, IndexBuffer.RESTART });
        // 3267 +y
        normal = new Vector3f(0.0f, 1.0f, 0.0f);
        positions[3].store(vertices); normal.store(vertices); new Vector2f(0.0f, 0.0f).store(vertices);
        positions[2].store(vertices); normal.store(vertices); new Vector2f(1.0f, 0.0f).store(vertices);
        positions[7].store(vertices); normal.store(vertices); new Vector2f(0.0f, 1.0f).store(vertices);
        positions[6].store(vertices); normal.store(vertices); new Vector2f(1.0f, 1.0f).store(vertices);
        indices.put(new int[] { 20, 21, 22, 23, IndexBuffer.RESTART });

        vertices.position(0);
        indices.position(0);
        return createGeometry(indices, IndexBuffer.TRIANGLE_STRIP, vertices);
    }
    
    public static Geometry createSphere(int subDivide, float radius, float posX, float posY, float posZ) {
        FloatBuffer vertices = BufferUtils.createFloatBuffer(8 * (subDivide + 1) * (subDivide / 2 + 1));
        IntBuffer indices = BufferUtils.createIntBuffer((2 * subDivide + 2 + 1) * (subDivide / 2));
        float theta = 0.0f;
        float dPhi = MathUtil.PI_MUL2 / (float)subDivide;
        float dTheta = MathUtil.PI / (float)(subDivide / 2);
        for(int iTheta=0; iTheta <= subDivide / 2; ++iTheta, theta += dTheta) {
            float phi = 0.0f;
            for(int iPhi=0; iPhi <= subDivide; ++iPhi, phi += dPhi) {
                float sinTheta = MathUtil.sin(theta);
                float cosTheta = MathUtil.cos(theta);
                float sinPhi = MathUtil.sin(phi);
                float cosPhi = MathUtil.cos(phi);
                Vector3f normal = new Vector3f(sinTheta * cosPhi, sinTheta * sinPhi, cosTheta);
                new Vector3f(posX + radius * normal.x, posY + radius * normal.y, posZ + radius * normal.z).store(vertices);
                normal.store(vertices);
                new Vector2f(phi / MathUtil.PI_MUL2, theta / MathUtil.PI).store(vertices);
                
                if(iTheta != subDivide / 2) {
                    indices.put((iTheta + 1) * (subDivide + 1) + iPhi);
                    indices.put(iTheta * (subDivide + 1) + iPhi);
                }
            }
            if(iTheta != subDivide / 2) {
                indices.put(IndexBuffer.RESTART);
            }
        }
        indices.position(0);
        vertices.position(0);
        return createGeometry(indices, IndexBuffer.TRIANGLE_STRIP, vertices);
    }
    
    public static Geometry buildFromObjData(ObjData data, boolean mirrored) {
    	
    	Geometry geo = geometries.get(data.getResource());
    	if (geo == null) { 
    		geo = new Geometry();
    	} else {
    		System.out.println("RETURN!!!!!!");
    		return geo;
    	}
    	
    	IndexBuffer indexBuffer = new IndexBuffer();
        if(!indexBuffer.init()) {
            return null;
        }
        VertexBuffer vertexBuffer = new VertexBuffer();
        if(!vertexBuffer.init()) {
            indexBuffer.destroy();
            return null;
        }
        vertexBuffer.addAttribute(VertexBuffer.ATTR_POS, 3);
        vertexBuffer.addAttribute(VertexBuffer.ATTR_NORMAL, 3);
        vertexBuffer.addAttribute(VertexBuffer.ATTR_TEX, 2);
        
        List<Float> vertices = new LinkedList<>();
        List<Integer> indices = new LinkedList<>();
        Map<ObjData.ObjIndex, Integer> indexMappings = new HashMap<>();
        for(Material material : data.getMaterials()) {
            for(ObjData.Face face : data.getFacesByMaterial(material)) {
                for(int i = 0; i < face.getIndexCount(); ++i) {
                    ObjData.ObjIndex objIndex = face.getIndex(mirrored ? i : face.getIndexCount() - 1 - i);
                    Integer index = indexMappings.get(objIndex);
                    if(index == null) {
                        Vector3f vertex = data.getVertex(objIndex.getVertex());
                        Vector3f normal = data.getNormal(objIndex.getNormal());
                        Vector2f tex = data.getTexture(objIndex.getTex());
                        index = vertices.size() / 8;
                        vertices.add(mirrored ? -vertex.x : vertex.x);
                        vertices.add(vertex.y);
                        vertices.add(vertex.z);
                        vertices.add(mirrored ? -normal.x : normal.x);
                        vertices.add(normal.y);
                        vertices.add(normal.z);
                        vertices.add(tex.x);
                        vertices.add(tex.y);
                        indexMappings.put(objIndex, index);
                    }
                    indices.add(index);
                }
                indices.add(IndexBuffer.RESTART);
            }
        }
        IntBuffer indexIntBuffer = BufferUtils.createIntBuffer(indices.size());
        for(Integer index : indices) {
            indexIntBuffer.put(index);
        }
        indexIntBuffer.position(0);
        indexBuffer.setData(indexIntBuffer, IndexBuffer.TRIANGLE_FAN);
        FloatBuffer vertexFloatBuffer = BufferUtils.createFloatBuffer(vertices.size());
        for(Float f : vertices) {
            vertexFloatBuffer.put(f);
        }
        vertexFloatBuffer.position(0);
        vertexBuffer.setData(vertexFloatBuffer);
        boolean success = geo.init(indexBuffer, vertexBuffer);
        if(!success) {
            vertexBuffer.destroy();
            indexBuffer.destroy();
            
            return null;
        } else {
        	geometries.put(data.getResource(), geo);
        	return geo;
        }
        
    }
    
    /**
     * Erzeugt eine fertig initialisierte Geometrie aus Index- und
     * Vertexbuffern.
     * @param indices Der Indexbuffer
     * @param topology Die Topologie des Indexbuffers
     * @param vertices Der Vertexbuffer
     * @return Die Geometrie oder <tt>null</tt> bei Fehlern
     */
    private static Geometry createGeometry(IntBuffer indices, int topology, FloatBuffer vertices) {
        VertexBuffer vertexBuffer = new VertexBuffer();
        IndexBuffer indexBuffer = new IndexBuffer();
        Geometry array = new Geometry();
        if(vertexBuffer.init() && indexBuffer.init()) {
            indexBuffer.setData(indices, topology);
            vertexBuffer.setData(vertices);
            vertexBuffer.addAttribute(VertexBuffer.ATTR_POS, 3);                // TODO!
            vertexBuffer.addAttribute(VertexBuffer.ATTR_NORMAL, 3);             // TODO!
            vertexBuffer.addAttribute(VertexBuffer.ATTR_TEX, 2);
            if(array.init(indexBuffer, vertexBuffer)) {
                return array;
            }
        }
        indexBuffer.destroy();
        vertexBuffer.destroy();
        return null;
    }
    
    /**
     * Erzeugt eine fertig initialisierte Geometrie aus Index- und Vertexdaten.
     * @param rawIndices Die Indizes
     * @param topology Die Topologie der Indexdaten
     * @param rawVertices Die Vertices
     * @return Die Geometrie oder <tt>null</tt> bei Fehlern
     */
    private static Geometry createGeometry(int rawIndices[], int topology, float rawVertices[]) {
        FloatBuffer vertices = BufferUtils.createFloatBuffer(rawVertices.length);
        vertices.put(rawVertices);
        vertices.flip();
        IntBuffer indices = BufferUtils.createIntBuffer(rawIndices.length);
        indices.put(rawIndices);
        indices.flip();
        return createGeometry(indices, topology, vertices);
    }
}
