/*
 * This software is the proprietary information of nico3000.
 */
package view.scene;

import actor.Actor;
import actor.ActorFactory;
import actor.components.RenderComponent;
import event.EventData;
import event.EventListener;
import event.events.ActorCreatedEvent;
import event.events.ActorDeletedEvent;
import event.events.ActorEvent;
import graphics.shader.Effect;
import logging.Logger;
import main.GameApp;
import mesh.Material;
import view.ScreenElement;
import view.scene.nodes.DrawableNode;
import view.scene.nodes.RootNode;
import view.scene.nodes.SceneNode;

/**
 * Ein <tt>ScreenElement</tt>, das einen Szenegraphen verwaltet.
 * @author nico3000
 */
public class SceneScreenElement extends ScreenElement implements EventListener {
    /**
     * Der Name der Szene.
     */
    private final String name;

    /**
     * Konstruktor.
     * @param name Der Name der Szene
     */
    public SceneScreenElement(String name) {
        this.name = name;
    }

    /**
     * Liefert den Namen der Szene.
     * @return Der Name der Szene
     */
    public String getName() {
        return this.name;
    }

    /**
     * Der Wurzelknoten des Szenegraphen
     */
    private SceneNode root = new RootNode();
    /**
     * Ein Matrixstack, um das Rendern zu vereinfachen.
     */
    private final MatrixStack stack = new MatrixStack();
    /**
     * Der aktuelle Effekt.
     */
    private Effect effect = new Effect();
    /**
     * Die aktuelle primaere Kamera.
     */
    private PerspectiveCamera camera;
    
    /**
     * Fuegt dem Szenegraphen einen Knoten hinzu. Dazu wird der Vaterknoten
     * anhand der ID des Actors identifiziert und diesem Knoten hinzugefuegt.
     * Wird die ID des Vaters als <tt>ActorFactory.INVALID_ACTOR_ID</tt>
     * angegeben, so wird der Knoten dem Wurzelknoten hinzugefuegt.
     * @param fatherActorID Die ID des Vaterknotens
     * @param child Der Kindknoten
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean addChild(int fatherActorID, SceneNode child) {
        if(fatherActorID == ActorFactory.INVALID_ACTOR_ID) {
            return this.root.addChild(child);
        } else {
            SceneNode node = this.root.getNode(fatherActorID);
            if(node == null) {
                Logger.INSTANCE.warning("unknown father for scnenenode");
                return this.root.addChild(child);
            } else {
                //System.err.println(node.getActorId() + " -> " + child.getActorId());
                return node.addChild(child);
            }
        }
    }
    
    /**
     * Entfernt einen Knoten aus dem Szenegraphen anhand einer ActorID.
     * @param actorId Die ID des Actors
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean removeChild(int actorId) {
        return this.root.removeChild(actorId);
    }

    /**
     * Liefert den Matrixstack.
     * @return Der Matrixstack
     */
    public MatrixStack getStack() {
        return this.stack;
    }

    @Override
    public void trigger(EventData data) {
        switch(data.getId()) {
            case ActorCreatedEvent.ID:
                Actor actor = GameApp.getGameLogic().getActor(((ActorCreatedEvent)data).getActorId());
                if(actor == null) {
                    Logger.INSTANCE.error("actor does not exist");
                    return;
                }
                if(actor.getComponent(RenderComponent.ID) != null) {
                    DrawableNode node = new DrawableNode();
                    if(node.init(actor.getId())) {
                        if(!this.addChild(GameApp.getGameLogic().getActorId(actor.getParentActorName()), node)) {
                            Logger.INSTANCE.error("DrawableNode for actor " + actor.getName() + " colud not be added");
                            node.destroy();
                        }
                    } else {
                        Logger.INSTANCE.error("DrawableNode for actor " + actor.getName() + " colud not be initialized");
                    }
                }                
                break;
            case ActorDeletedEvent.ID:
                this.removeChild(((ActorEvent)data).getActorId());
                break;
        }
    }

    /**
     * Setzt die aktuelle Kamera der Szene.
     * @param camera Die aktuelle Kamera
     */
    public void setCamera(PerspectiveCamera camera) {
        this.camera = camera;
    }

    /**
     * Liefert die aktuelle Kamera der Szene.
     * @return Die aktuelle Kamera
     */
    public PerspectiveCamera getCamera() {
        return this.camera;
    }

    @Override
    public boolean onRestore() {
        this.root.restore();
        this.effect.initVF("Shaders/DefaultVS.glsl", "Shaders/DefaultFS.glsl");
        Material.modifyEffectToUseMaterial(this.effect);
        return true;
    }

    @Override
    public boolean onLostDevice() {
        this.root.lostDevice();
        return true;
    }

    @Override
    public void onRender() {
        this.effect.use();
        if(this.camera != null) {
            this.camera.bindBuffer();
        }
        this.root.preRender(this);
        this.root.render(this);
        this.root.renderChildren(this);
        this.root.postRender(this);
        if(!this.stack.isEmpty()) {
            Logger.INSTANCE.error("Matrix stack is not empty!");
            this.stack.reset();
        }
    }

    @Override
    public void onUpdate(long deltaMillis) {
        this.root.update(this, deltaMillis);
    }

    @Override
    public void onDestroy() {
        this.effect.destroy();
        this.root.destroy();
    }    
}
