/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package view.scene.nodes;

import actor.Actor;
import actor.components.RenderComponent;
import actor.components.TransformComponent;
import event.EventData;
import event.EventListener;
import event.events.ActorMovedEvent;
import graphics.shader.UniformBuffer;
import java.nio.FloatBuffer;
import java.util.LinkedList;
import java.util.List;
import logging.Logger;
import main.GameApp;
import mesh.BoundingSphere;
import org.lwjgl.BufferUtils;
import util.Pose;
import view.scene.SceneScreenElement;

/**
 * Ein Node zu einem Actor, der gezeichnet werden kann.
 * @author nico3000
 */
public class DrawableNode extends SceneNode implements EventListener {
    /**
     * Die Pose des Actors.
     */
    private Pose pose;
    /**
     * Die Renderkomponente des Actors.
     */
    private RenderComponent renderComponent;
    /**
     * Die Model- und inverse Modelmatrix des Actors.
     */
    private UniformBuffer modelBuffer = new UniformBuffer();
    /**
     * Der Floatbuffer, der die Model- und inverse Modelmatrix enthaelt.
     */
    private final FloatBuffer modelBufferData = BufferUtils.createFloatBuffer(32);
    /**
     * Hilfsliste, die alle BoundingSpheres der direkten Nachfahren und die
     * dieses Knotens enthaelt.
     */
    private final List<BoundingSphere> spheres = new LinkedList<>();
    /**
     * Die Seite des Frustums, bei der der letzte Sichtbarkeitstest
     * fehlgeschlagen ist. Der naechste Test startet bei diese Seite, da es sehr
     * wahrscheinlich ist, dass im naechsten Frame dieselbe Seite fuer ein
     * Scheitern verantwortlich sein wird.
     */
    private int lastFrustumSide = 0;
    /**
     * Die BoundingSphere in Weltkoordinaten, die die BoundingSphere dieses
     * Knotens und aller direkten Nachfahren enthaelt.
     */
    private BoundingSphere boundingSphere = new BoundingSphere();
    /**
     * Gibt an, ob sich die BoundingSphere seit dem letzten <tt>render()</tt>
     * Aufruf veraendert hat.
     */
    private boolean boundingSphereChanged = false;

    /**
     * Initialisiert den Knoten mit einem Actor. Der Actor muss mindestens die
     * TransformComponent und RenderComponent enthalten.
     * @param actorId Die ID des Actors
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init(int actorId) {
        Actor actor = GameApp.getGameLogic().getActor(actorId);
        if(actor == null) {
            Logger.INSTANCE.error("invalid actor id");
            return false;
        }
        TransformComponent transformComponent = (TransformComponent)actor.getComponent(TransformComponent.ID);
        if(transformComponent == null) {
            Logger.INSTANCE.error("actor has no transform component");
            return false;
        }
        this.pose = transformComponent.getPose();
        this.renderComponent = (RenderComponent)actor.getComponent(RenderComponent.ID);
        if(this.renderComponent == null) {
            Logger.INSTANCE.error("actor has no render component");
            this.pose = null;
            return false;
        }
        if(!this.modelBuffer.init()) {
            Logger.INSTANCE.error("model buffer could not be initialized");
            this.renderComponent = null;
            this.pose = null;
            return false;
        }
        this.modelBuffer.setBindingIndex(UniformBuffer.BINDING_MODEL);
        GameApp.getEventManager().register(this, ActorMovedEvent.ID);
        super.setActorId(actorId);
        return true;
    }

    @Override
    public void setActorId(int actorId) {
        Logger.INSTANCE.error("cannot set actor id of DrawableNode, use init() instead");
    }

    @Override
    public void preRender(SceneScreenElement scene) {   
        scene.getStack().push(this.pose.getModel(), this.pose.getModelInv());
        if(this.hasMoved) {
            this.modelBufferData.position(0);
            scene.getStack().topMat().store(this.modelBufferData);
            scene.getStack().topMatInverse().store(this.modelBufferData);
            this.modelBufferData.position(0);
            this.modelBuffer.setData(this.modelBufferData);
            this.hasMoved = false;
        }
        if(this.boundingSphereChanged) {
            this.boundingSphere.transform(scene.getStack().topMat());
            this.boundingSphereChanged = false;
        }
        this.modelBuffer.bind();        
    }

    @Override
    public void render(SceneScreenElement scene) {
        this.renderComponent.drawGeometries();
    }

    @Override
    public void postRender(SceneScreenElement scene) {
        scene.getStack().pop();
    }

    @Override
    public void trigger(EventData data) {
        switch(data.getId()) {
            case ActorMovedEvent.ID:
                if(this.getActorId() == ((ActorMovedEvent)data).getActorId()) {
                    this.actorMoved();
                } else {
                }
                break;
        }
    }

    @Override
    public void destroy() {
        super.destroy();
        GameApp.getEventManager().unregister(this, ActorMovedEvent.ID);
        this.modelBuffer.destroy();
    }

    @Override
    public boolean isVisible(SceneScreenElement scene) {
        if(scene.getCamera() == null) {
            return false;
        } else {
            for(int i=0; i < 6; ++i) {
                int side = (this.lastFrustumSide + i) % 6;
                if(this.boundingSphere.isCompleteOnInvisibleSide(scene.getCamera().getFrustumFaces()[side])) {
                    this.lastFrustumSide = side;
                    return false;
                }
            }
            return true;
        }
    }
    
    @Override
    protected void successorMoved() {
        this.spheres.clear();
        if(this.renderComponent.getBoundingSphere() != null) {
            this.spheres.add(this.renderComponent.getBoundingSphere());
        }
        for(SceneNode child : this.getChildren()) {
            if(child instanceof DrawableNode) {
                this.spheres.add(((DrawableNode)child).boundingSphere);
            }
        }
        this.boundingSphere.encapsulateSpheres(this.spheres);
        this.boundingSphere.transform(this.pose.getModel());
        this.boundingSphereChanged = true;
        super.successorMoved();
    }
}
